/*+
    Name:	vgatest.c
    Date:	16-Aug-1988
    Author:	Kent J. Quirk
    Abstract:	This program tests VGA compatibility.
-*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <bios.h>
#include <stdarg.h>
#include <conio.h>
#include <ctype.h>
#include <setjmp.h>
#include "vgabios.h"
#include "vgatest.h"
#include "vgaregs.h"

typedef struct {
    int off;
    int nbytes;
    char *name;
} COMPARES;

COMPARES compare_general[] = {
     0, 1, "Misc out",
     1, 1, "Feat Ctrl",
     2, 1, "Input Status 0",
    -1, 1, NULL
};

COMPARES compare_sequencer[] = {
     0, 1, "Clk Mode", 
     1, 1, "Map Mask", 
     2, 1, "Char Map",
     3, 1, "Mem Mode",
    -1, 1, NULL
};    

COMPARES compare_dac[] = {	    /* a bit of a kludge because it seemed */
    -1, 1, NULL 		       /* silly to write out 256 regs */
};

COMPARES compare_crtc[] = {
     0, 1, "h_total",
     1, 1, "h_disp_end",
     2, 1, "h_blank_start",
     3, 1, "h_blank_end",
     4, 1, "h_retrace_start",
     5, 1, "h_retrace_end",
     6, 1, "v_total",
     7, 1, "overflow",
     8, 1, "preset_row_scan",
     9, 1, "max_scan_line",
    10, 1, "cur_start",
    11, 1, "cur_end",
    12, 1, "start_addr_hi",
    13, 1, "start_addr_lo",
    14, 1, "curs_loc_hi",
    15, 1, "curs_loc_lo",
    16, 1, "v_retrace_start",
    17, 1, "v_retrace_end",
    18, 1, "v_display_end",
    19, 1, "offset",
    20, 1, "underline_loc",
    21, 1, "v_blank_start",
    22, 1, "v_blank_end",
    23, 1, "mode_control",
    24, 1, "line_compare",
    -1, 1, NULL
};    

COMPARES compare_attrib[] = {
     0, 1, "palette",
     1, 1, "mode_control",
     2, 1, "overscan",
     3, 1, "color_plane_enable",
     4, 1, "h_pel_panning",
     5, 1, "color_select",
    -1, 1,  NULL
};    

COMPARES compare_dyn_state[] = {
    0x04, 1, "display mode",
    0x05, 2, "# columns",
    0x07, 2, "buf length",
    0x09, 2, "start addr",
    0x1e, 2, "CRT Ctrl addr",
    0x20, 1, "CRT_Mode_Set",
    0x21, 1, "CRT_Palette",
    0x22, 1, "# char rows",
    0x23, 2, "scanlines/char",
    0x25, 1, "Active DCC",
    0x26, 1, "Alt DCC",
    0x27, 2, "# colors",
    0x29, 1, "# pages",
    0x2A, 1, "# scan lines index",
    0x2B, 1, "primary active char block",
    0x2C, 1, "secondary active char block",
    0x2D, 1, "misc state info",
    0x31, 1, "vid memory index",
    0x32, 1, "save ptr state",
    -1, 1,  NULL
};

typedef struct {
    COMPARES *array;
    char *filename;
    char *regset;
    int datasize;
} COMPARE_ITEM;

COMPARE_ITEM cmp[] = {
    { compare_general,	 "GENERAL",  "General", sizeof(GENERAL_REGS) },
    { compare_dac,	 "DAC",      "DAC Color Registers", sizeof(DAC_REGS) },
    { compare_sequencer, "SEQUENCE", "Sequencer", sizeof(SEQUENCER_REGS) },
    { compare_crtc,	 "CRTC",     "CRTC", sizeof(CRTC_REGS) },
    { compare_attrib,	 "ATTRIB",   "Attribute", sizeof(ATTRIB_REGS) },
    { compare_dyn_state, "DSTATE",   "Dynamic State", sizeof(STATE) },
};

typedef struct {
    void (*func)();
    char *name;
} TEST;

TEST tests[] = {
    { test_mode_support, "BIOS: Mode Support" },
    { test_cursor_operation, "BIOS: Cursor Operation" },
    { test_light_pen, "BIOS: Light Pen Support" },
    { test_multiple_pages, "BIOS: Multiple Pages" }, 
    { test_screen_scrolling, "BIOS: Screen Scrolling" },
    { test_text_io, "BIOS: Text Input/Output" },
    { test_graphics_io, "BIOS: Graphics Input/Output" },
    { test_palette_functions, "BIOS: Palette/DAC" },
    { test_video_save_restore, "BIOS: Save/Restore Video State" },
    { test_character_generator, "BIOS: Character Generator" },
    { test_alternate_select, "BIOS: Alternate Select" },
    { test_bios_data, "BIOS: Data Area Compatibility" },
    { test_general_regs, "Hardware: General Registers" },
    { test_dac_regs, "Hardware: DAC Registers" },
    { test_sequencer_regs, "Hardware: Sequencer Registers" },
    { test_crtc_regs, "Hardware: CRTC Registers" },
    { test_graphics_regs, "Hardware: Graphics Registers" },
    { test_attribute_regs, "Hardware: Attribute Registers" },
};

#define hbound(a) (sizeof(a)/sizeof(a[0]))
#define ERRORMAX 50		/* max # of errors before it quits a test */

FILE *msgfile;
extern VIDSTAT savemode;
int errorcount = 0;
int vidmode = COLORVID;
jmp_buf jump_mark;
int write_data = 0;
static int testnumber = 0;
int fast = 0;

/**** p a u s e k e y ****
    Abstract:	Pauses for a given number of ticks, or terminates
		early if the user presses a key.
    Parameters: The number of ticks to pause (at 18.2 ticks/second).
		If nticks==2, calibrates the quarter-tick timer.
		If nticks==0, pauses for one quarter of a tick.
    Returns:	The key pressed, if one was pressed, or -1 if it timed out.
		If the user ever presses ESCape, this will always return
		immediately with ESC as the keystroke.	This allows inner
		loops to call this routine without panicking about the return
		value.
    Comments:	Will never time out if the pause period will cross midnight
		(but pressing a key will still work).
		Might be short by a full tick (will anyone notice?).
****************************/
int pausekey(int nticks)
{
    int c;
    static int escape = 0;
    static long quickcount = 0;
    long waitfor, current, counter = quickcount / 4;
    
    if (escape)
	return(ESC);
	
    if (fast)
	nticks = 0;	    /* always return quickly */
    
    _bios_timeofday(_TIME_GETCLOCK, &current);
    waitfor = current + nticks;
    if (nticks == 0)
	waitfor += 10;
    while (waitfor >= current)
    {
	_bios_timeofday(_TIME_GETCLOCK, &current);
	if (kbhit())
	{    
	    escape = ((c = getch()) == ESC);
	    return(c);
	}
	if (nticks == 2)	    /* calibrate counter */
	{
	    if ((waitfor - current) == 1)
		++quickcount;
	}
	if (nticks == 0)	    /* run a short count */
	{
	    if (--counter <= 0)
		break;
	}
    }
    return(-1);
}

/**** e r r o r ****
    Abstract:	Takes parameters identical to printf, but prints them 
		on msgfile (since we're playing with video output).
    Parameters: See printf documentation.
    Returns:	Whatever printf does.
****************************/
int error(char *format,...)
{
    va_list arglist;
    int retval;
    static int lasttest = -1;
    static int thiserror = 0;

    if (testnumber != lasttest)
    {	 
	++errorcount;
	thiserror = 0;
	lasttest = testnumber;
	fprintf(msgfile, "Test %2d:  %s\n", testnumber+1, 
		tests[testnumber].name);
    }

    fprintf(msgfile, "  ");
    va_start(arglist, format);		/* start argument list */
    retval = vfprintf(msgfile, format, arglist);
    va_end(arglist);
    if (++thiserror > ERRORMAX)
    {
        SetMode(savemode.video_mode);
        fprintf(msgfile, "  Aborting test %2d after %d errors.\n", 
                    testnumber+1, ERRORMAX);
	longjmp(jump_mark, 0);
    }
    return(retval);
}

/**** c o m p a r e _ d a t a ****
    Abstract:
    Parameters:
    Returns:
    Comments:
****************************/
void compare_data(int ci, int mi, void *data)
{
    FILE *f;
    char filename[20];
    void *fdata;
    int i, j;
	      
    fdata = malloc(cmp[ci].datasize);
    
    strcpy(filename, cmp[ci].filename);
    if (vidmode == COLORVID)
	strcat(filename, ".COL");
    else
	strcat(filename, ".MON");
    
    if (write_data)
    {
	if ((f = fopen(filename, "rb+")) != NULL)
	{
	    fseek(f, (long)(mi * cmp[ci].datasize), SEEK_SET);
	    fwrite(data, 1, cmp[ci].datasize, f);
	    fclose(f);
	}
    }

    if ((f = fopen(filename, "rb")) == NULL)
    {	 
	error("Register file '%s' not found.\n", filename);
	return;
    }
    fseek(f, (long)(mi * cmp[ci].datasize), SEEK_SET);
    if (fread(fdata, 1, cmp[ci].datasize, f) != cmp[ci].datasize)
    {
	error("Unable to read data for mode %02X from '%s'.\n", 
	    valid_modes[mi].video_mode, filename);
	return;
    }
    fclose(f);
    
    for (i=0; cmp[ci].array[i].off != -1; i++)
    {	 
	if (memcmp((BYTE *) data + cmp[ci].array[i].off,
		   (BYTE *)fdata + cmp[ci].array[i].off,
		   cmp[ci].array[i].nbytes) != 0)
	{
	    switch (cmp[ci].array[i].nbytes) {
	    case 1:
		error("Mode %02Xh, %s -- %s: IBM=%02Xh   Current=%02Xh\n",
		    valid_modes[mi].video_mode,
		    cmp[ci].regset,
		    cmp[ci].array[i].name,
		    ((BYTE *)fdata)[cmp[ci].array[i].off],
		    ((BYTE *) data)[cmp[ci].array[i].off] );
		break;
	    case 2:
		error("Mode %02Xh, %s -- %s: IBM=%04Xh Current=%04Xh\n",
		    valid_modes[mi].video_mode,
		    cmp[ci].regset,
		    cmp[ci].array[i].name,
		    *((WORD *)((BYTE *)fdata+cmp[ci].array[i].off)),
		    *((WORD *)((BYTE *) data+cmp[ci].array[i].off)) );
		break;
	    }
	}
    }
    
    if (i == 0) 	/* we didn't do anything, so compare DAC data */
    {
	if (memcmp(data, fdata, cmp[ci].datasize) != 0)
	{    
	    DAC_COLOR *ibm, *cur;
	    
	    error("%s -- IBM and current do not agree.\n", cmp[ci].regset);
	    ibm = (DAC_COLOR *)fdata;
	    cur = (DAC_COLOR *) data;
	    for (j=0; j<256; j++)
	    {
		if (memcmp(ibm+j, cur+j, sizeof(DAC_COLOR)) != 0)
		    error("  Color %d: IBM=(%02X%02X%02X), Current=(%02X%02X%02X)\n",
			j, ibm[j].red, ibm[j].blu, ibm[j].grn,
			   cur[j].red, cur[j].blu, cur[j].grn);
	    }
	}
    }
    free(fdata);
}

/**** p r i n t _ g r e e t i n g ****
    Abstract:   Prints the greeting screen, gives the user a chance to back
                out.
    Parameters: None.
    Returns:    Nothing, or never returns if user presses ESC.
****************************/
void print_greeting()
{
    static char *greet_text[] = {
        "\n\n\n\n\n\n\n                                PC Tech Journal\n\n", 
        "                            VGA Compatibility Test\n\n",
        "                                  Version 1.0\n\n\n\n",
        "                     PRESS ENTER TO CONTINUE, ESC TO QUIT\n\n\n\n",
        "                (c) Copyright 1988 Ziff Communications Company\n\n\n\n",
        NULL
    };
    int i;
   
    for (i=0; greet_text[i] != NULL; i++)
        printf("%s", greet_text[i]);
    for (;;)
    {
        switch(getch()) {
        case ESC:                    /* escape key */
            exit(1);
            break;
        case '\n':
        case '\r':
            return;
            break;
        default:
            printf("\a");
            break;
        }
    }
}

/**** u s a g e ****
    Abstract:   Prints a usage and copyright message.
    Parameters: None.
    Returns:	Never returns.
    Comments:
****************************/
void usage()
{   
    int i;
    static char *usage_text[] = {
    "\n                                PC Tech Journal\n",
    "\n                            VGA Compatibility Test",
    "\n                                  Version 1.0",
    "\n                        by Kent Quirk and Ed McNierney\n",
    "\n\n                (c) Copyright 1988 Ziff Communications Company\n",
    "\n\n Usage:  VGATEST ERRORFILE [-tTEST#] [-f] [-h] [-?]\n",
    "\n         ERRORFILE is the file to which error messages are written \n",
    "         (they are not displayed).  This parameter must be specified.\n",
    "\n         TEST# is a single test number to execute.  If none is\n",
    "         specified, all tests are run.\n",
    "\n         -f (fast) skips pauses in the tests.\n",
    "\n         -h and -? print a list of all tests and their test numbers.\n\n",
    NULL
    };
   
    for (i=0; usage_text[i] != NULL; i++)
        printf("%s", usage_text[i]);
    exit(1);
}

void morefile(char *filename, int nlines)
{
    FILE *textfile;
    int c, line_ctr = 0;
    char buf[100];
    
    if ((textfile = fopen(filename, "rb")) == NULL)
    {
        printf("Couldn't open '%s'\n", filename);
        return;
    }
    
    while (!feof(textfile))
    {
        if (fgets(buf, sizeof(buf), textfile) == NULL)
        {
            fclose(textfile);
            return;
        }
        printf("%s", buf);
        if (line_ctr++ >= nlines)
        {
            printf("--MORE-- (ENTER=1 line, ESC=quit, any key=next screen");
            c = getch();
            printf("\r                                                      \r");
            switch (c) {
            case '\n':
            case '\r':
                break;
            case ESC:
                printf("\n");
                fclose(textfile);
                return;
            default:
                line_ctr = 0;
                break;
            }
        }
    }
    fclose(textfile);
    return;
}

/**** m a i n ****
    Abstract:	This is the main module for the program.  It lets the user
		select one or all tests, and then runs them.
    Parameters: The usual argc/argv set.
    Returns:	An errorlevel corresponding to the number of errors found.
		0 = none, 1 = one, 2 = two or more
****************************/
main(int argc, char *argv[])
{
    static char *filename = NULL;
    int c, i, j;
    int test=-1;	    /* run all tests */
    char *termmsg;
    
    for (i=1; i<argc; i++)
    {
	if (argv[i][0] == '-' || argv[i][0] == '/')
	{   
	    switch (argv[i][1]) 
	    {
                case 't': case 'T':
		    test = atoi(argv[i]+2) - 1;
		    break;
                case 'h': case 'H':
		case '?':
		    for (j=0; j<hbound(tests); j++)
			printf("%2d) %s\n", j+1, tests[j].name);
		    printf("Type VGATEST for command line options.\n");
		    exit(1);
		    break;
                case 'w': case 'W':
		    write_data = 1;
		    break;
                case 'f': case 'F':
		    fast = 1;		    /* doesn't pause if this is set */
		    break;
		default:
		    usage();
		    break;
	    }
	}
	else 
	    filename = argv[i];
    }

    if (filename == NULL)
	usage();
    
    print_greeting();
	
    if ((msgfile = fopen(filename, "w")) == NULL)
    {
	printf("Unable to open output file '%s'\n", filename);
	exit(1);
    }
    
    fprintf(msgfile, "VGA Test file %s\n", filename);
    
    GetStatus(&savemode);
    if ((read_general_reg(MISC_OUT_R) & 1) == 0)
	vidmode = MONOVID;	/* tell world about monochrome monitor */
    if (test == -1)
    {
	i = 0;
	if (setjmp(jump_mark) != 0)
	    i++;	/* make sure we do the increment we skipped */
	for (; i<hbound(tests); i++)	     /* run all tests */
	{    
	    testnumber = i;
	    (*tests[i].func)();
	    SetMode(savemode.video_mode);
	    if (pausekey(0) == ESC)
		break;
	}
    }
    else
    {
	testnumber = test;
	if (setjmp(jump_mark) == 0)
	    (*tests[test].func)();		/* run one test */
    }
    
    switch (errorcount) {
    case 0:
	termmsg = "There were no errors.\n";
	break;
    case 1:
	termmsg = "There was 1 test that had errors.\n";
	break;
    default:
	termmsg = "There were %d tests that had errors.\n";
	break;
    }

    printf(termmsg, errorcount);
    fprintf(msgfile, termmsg, errorcount);
    fclose(msgfile);
    printf("The test results were written to the file '%s'\n", filename);
    printf("Press ENTER to read them now, or ESC to return to DOS.");
    while ((c = getch()) != ESC)
    {
        if ((c == '\n') || (c == '\r'))
        {
            printf("\n");
            morefile(filename, 21);
            break;
        }
    }
    if (errorcount > 2)
	errorcount = 2;
    return(errorcount);
}
